// Copyright 2003-2012 Intel Corporation. All Rights Reserved.
// 
// The source code contained or described herein and all documents related 
// to the source code ("Material") are owned by Intel Corporation or its
// suppliers or licensors.  Title to the Material remains with Intel Corporation
// or its suppliers and licensors.  The Material is protected by worldwide
// copyright and trade secret laws and treaty provisions.  No part of the
// Material may be used, copied, reproduced, modified, published, uploaded,
// posted, transmitted, distributed, or disclosed in any way without Intel's
// prior express written permission.
// 
// No license under any patent, copyright, trade secret or other intellectual
// property right is granted to or conferred upon you by disclosure or delivery
// of the Materials, either expressly, by implication, inducement, estoppel
// or otherwise.  Any license under such intellectual property rights must
// be express and approved by Intel in writing.


#define BOOL bool
#include <stdlib.h>
#include "2DArray.h"
#include "common.h"
#include "life.h"

#include "tbb/task_scheduler_init.h"
#include "tbb/blocked_range2d.h"
#include "tbb/parallel_for.h"
#include <omp.h>

double secnds()
{
  return omp_get_wtime();
}

using namespace tbb;

Grid map0, map1;     // square array holding cells
int maxrow, maxcol;      // user defined grid size
char *fName = NULL;      // pointer to file name string from argv


int main(int argc, char *argv[])
{
   int i, gencount=0;
   double start, time;

   if (argc == 2) fName = argv[1];

printf("FILE NAME: %s\n\n",fName);


   Initialize(&map0, &map1, fName);
   start = secnds();
      for(i=GENS-1; i>= 0; i--){
         gencount++;
         if (gencount%2) {
           ClearMap(map1);
           NextGen(map0, map1);
         }
         else {
           ClearMap(map0);
           NextGen(map1, map0);
         }
      }
    time = secnds() - start;
    printf("Time =  %f\n",(float) time);
   return 0;
}

class CompNextGen {
	Grid oldMap;
	Grid newMap;

public:
	CompNextGen (Grid omap, Grid nmap): oldMap(omap), newMap(nmap) {} 
	void operator()( const blocked_range2d<int, int>& r ) const {
		//int ncount;
		//State current;

		for (int row = r.rows().begin(); row < r.rows().end(); row++)
			for (int col = r.cols().begin(); col < r.cols().end(); col++) {
				State current = oldMap[row][col];
				int ncount = NeighborCount(oldMap, row, col);
				newMap[row][col] = CellStatus(current, ncount);
			}
	}
};


void NextGen(Grid omap, Grid nmap)
{
	/* */
   int row, col;
   int ncount;
   State current;

   for (row = 1; row <= maxrow; row++)
      for (col = 1; col <= maxcol; col++) {
         current = omap[row][col];
         ncount = NeighborCount(omap,row,col);
         nmap[row][col] = CellStatus(current,ncount);
      }
/*	  */
//   parallel_for (blocked_range2d<int, int>(1, maxrow+1, 10, 1, maxcol+1, 10), CompNextGen(omap, nmap));

}

int NeighborCount(Grid map, int row, int col)         
{
   int nbrrow,              // loop index for row of neighbor loops
       nbrcol;              // column loop index
   int count = 0;

   for (nbrrow = row-1; nbrrow <= row+1; nbrrow++)
      for (nbrcol = col-1; nbrcol <= col+1; nbrcol++)
         if (nbrrow != row || nbrcol != col) {  //skip cell itself
            if (map[nbrrow][nbrcol] == ALIVE) count++;
         }

   return count;
}
        


State CellStatus(State curr, int ncount)
{
    State status;

    switch(ncount) {
    case 0:
    case 1:
       status = DEAD;
       break;
    case 2:
       status = curr;
       break;
    case 3: 
       status = ALIVE;
       break;
    case 4:
    case 5:
    case 6:
    case 7:
    case 8:
       status = DEAD;
       break;
    default:
       Error("Impossible case in CellStatus.");
    }  // switch

    return status;
}


void Initialize(Grid *oldmap, Grid *newmap, char *fName)
{
   int row, col;

// Put out an initial message and establish the bounds for the grid
   FindSize(&maxrow, &maxcol);

   if (!(*oldmap = Allocate2DArray<State>(MAXROW+2, MAXCOL+2))) exit(1);
   *newmap = Allocate2DArray<State>(MAXROW+2, MAXCOL+2);

// Obtain the initial configuration
   ReadMap(*oldmap, fName);
   ClearMap(*newmap);

}

BOOL UserSaysYes(void)
{
   int c;
   printf(" (y/n?)");
   do {
      while((c=getchar()) == '\n') ;  // ignore newline character
      if (c == 'y' || c == 'Y' || c == 'n' || c == 'N')
         return (c == 'y' || c == 'Y');
      printf("Please respond by typing one of the letters y or n\n");
   } while(1);
}

void WriteMap(Grid map, int gc)
{
   int row, col;
   int i;

   putchar('\n');
   putchar('\n');
   for (i=0; i < (MAXCOL/2-7); i++) putchar(' ');
   printf("GENERATION #%d\n",gc);
   putchar('\n');
   for (row=1; row<=maxrow; row++){
      for (col=1; col<=maxcol; col++)
         if(map[row][col] == ALIVE)
            putchar('*');
         else
            putchar('-');
      putchar('\n');
   }
}

void FindSize(int *maxrow, int *maxcol)
{
   *maxrow = MAXROW;
   *maxcol = MAXCOL;

   printf("This program is a simulation of the game of Life\n"
          "The grid has a size of %d rows and "
          "%d columns\n", MAXROW, MAXCOL);
   printf("Grid coordinates are numbered starting with 1 up to the\n"
          "number of rows or columns.  Positions outside of the grid\n"
          "area are incapable of sustaining living cells.\n\n");
}

void ReadMap(Grid map, char *fName)
{
   int row, col;
   FILE *fp;

   for (row=0; row <= MAXROW+1; row++)
      for (col=0; col <= MAXCOL+1; col++)
         map[row][col] = DEAD;     // Set all cells empty, including the hedge

   if (fName == NULL) {
      printf("Set up initial cell configuration:\n");
      printf("On each line give a pair of coordinates for a living cell.\n"
             "Terminate the list with the special pair 0 0.\n");
      scanf("%d %d",&row, &col);
      while (row != 0 || col != 0) {
         if (row >= 1 && row <= MAXROW && col >= 1 && col <= MAXCOL) {
            map[row][col] = ALIVE;
         }
         else
            printf("Values are not within range.\n");
   
         scanf("%d %d",&row, &col);
      }
      while (getchar() != '\n') ;  // Discard remaining characters
   }

   else {    // read from file
      int good=0, bad=0, dup=0;
      if ((fp = fopen(fName, "r")) == NULL) Error("Input file not found");
      printf("Reading input data from %s...",fName); 
      fscanf(fp,"%d %d",&row, &col);
//printf(" %d %d \n",row, col);
      while (row != 0 || col != 0) {
         if (row >= 1 && row <= MAXROW && col >= 1 && col <= MAXCOL) {
            if (map[row][col] == DEAD) {
               good++;
            }
            else { 
               dup++;
            }
            map[row][col] = ALIVE;
         }
         else {
            bad++;
         }
         fscanf(fp,"%d %d",&row, &col);
//printf(" %d %d \n",row, col);
      }
      printf("Done.\n %d Cells created",good);
      if (bad > 0) printf (", %d Cells out of range",bad);
      if (dup > 0) printf ("and %d Cells duplicated",dup);
      printf("\n");
      fclose(fp);
   }
   
}

void ClearMap(Grid map)
{
   int row, col;

// Set all entries in map to 0
   for (row=0; row <= maxrow+1; row++)
      for (col=0; col <= maxcol+1; col++)
         map[row][col] = DEAD;
}
